import logging

# https://github.com/jedie/django-for-runners
from for_runners.services.gpx_calculate_values import calculate_values
from for_runners.services.gpx_save_gpx import save_gpx_file
from for_runners.services.gpx_svg_generator import generate_svg


log = logging.getLogger(__name__)


def gpx_pre_save_handler(sender, instance, update_fields, **kwargs):
    if instance.gpx:
        calculate_values(gpx_track=instance)


def gpx_post_save_handler(sender, instance, created, update_fields, **kwargs):
    if not instance.track_svg:

        generate_svg(gpx_track=instance, force=False)

        save_gpx_file(gpx_track=instance, force=False)

        log.warning("Save gpx file, too!")



#!/usr/bin/env python3

import os

import subprocess
import sys
import time


import click

# https://github.com/jedie/django-for-runners
from for_runners import __version__
from for_runners_project.utils.gunicorn_server import get_gunicorn_application


os.environ["DJANGO_SETTINGS_MODULE"] = "for_runners_project.settings"

# print("sys.real_prefix:", getattr(sys, "real_prefix", "-"))
# print("sys.prefix:", sys.prefix)

try:
    import django
    from django.core.management import call_command
except ImportError as err:
    print(f"\nERROR: {err}\n")
    import traceback

    traceback.print_exc()
    print("")
    print(" *** Couldn't import Django. ***")
    print(" *** Did you forget to activate a virtual environment? ***")
    print("")
    sys.exit(101)

if sys.version_info < (3, 5):
    print("\nERROR: Python 3.5 or greater is required!\n")
    sys.exit(101)


def verbose_subprocess(*args, **kwargs):
    """
    Call this script via subprocess, but with a "command" argument
    """
    print("\n" * 3)
    print("_" * 79, file=sys.stderr, flush=True)
    click.echo(f"subprocess call '{click.style(' '.join(args), fg='green')}'\n")
    subprocess.check_call(args, **kwargs)


def subprocess_manage(*args):
    """
    Call django manage command via:
        for_runners_project/__main__.py
    """
    args = [sys.executable, "-m", "for_runners_project"] + list(args)
    verbose_subprocess(*args)


def call_manage_command(*, cmd_class):
    cmd = cmd_class.Command()
    call_command(cmd)


@click.group()
@click.version_option(__version__, prog_name="Django-ForRunners")
def cli():
    print("Start up...")


@cli.command()
def run_dev_server():
    """
    run the django dev server in endless loop.

    entry points used in setup.py
    e.g.:
        ~$ ~/Django-ForRunners/bin/for_runners run-server
    """
    while True:
        try:
            subprocess_manage("makemigrations")  # helpfull for developing and add/change models ;)
            subprocess_manage("migrate")
            subprocess_manage("run_server")
        except Exception as err:
            print(f"\nError: {err}")
        except KeyboardInterrupt:
            pass

        print("")
        print("\nRestart, Quit with CONTROL-C")
        try:
            for no in range(5, 1, -1):
                print("\tin %i sec..." % no)
                time.sleep(1)
        except KeyboardInterrupt:
            print("\n\nExit by keyboard interrupt, ok.\n")
            sys.exit(0)


@cli.command()
@click.argument("address", nargs=1, default="127.0.0.1:8000")
def run_server(address):
    """
    Run the gunicorn server in endless loop.
    e.g.:

    ~$ for_runners run-server

    ~$ for_runners run-server 127.0.0.1:8080
    """
    subprocess_manage("makemigrations")  # helpfull for developing and add/change models ;)
    subprocess_manage("migrate")
    gunicorn_application = get_gunicorn_application(address)
    gunicorn_application.run()


@cli.command()
def create_starter():
    """
    Create starter file.
    """
    django.setup()
    from for_runners_project.starter import create_starter

    create_starter()


@cli.command()
def recreate_files():
    """
    Recreate all SVG and GPX files for all Tracks.
    """
    django.setup()
    from for_runners.management.commands import recreate_files

    call_manage_command(cmd_class=recreate_files)


@cli.command()
def backup():
    """
    Backup everything
    """
    django.setup()
    from for_runners.management.commands import backup

    call_manage_command(cmd_class=backup)


if __name__ == "__main__":
    cli()



"""
    created 31.05.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import sys
from pathlib import Path

from django.contrib.auth.models import User

# https://github.com/jedie/django-for-runners
from for_runners.management.commands.base import BaseCommand
from for_runners.models import GpxModel
from for_runners.services.gpx_create import add_from_files


class Command(BaseCommand):
    help = "Import GPS files (*.gpx)"

    def add_arguments(self, parser):
        parser.add_argument(
            "--username",
            "-u",
            action="store",
            dest="username",
            required=True,
            help="The user to assign to the imported files",
        )
        parser.add_argument("path", help="Path to *.gpx files")

    def handle(self, *args, **options):
        username = options.get("username")
        try:
            user = User.objects.get(username=username)
        except User.DoesNotExist as err:
            self.stderr.write(f"ERROR getting user {username!r}: {err}")
            user_names = ", ".join(User.objects.values_list("username", flat=True))
            self.stdout.write(f"Existing usernames are: {user_names}")
            sys.exit(3)

        self.stdout.write(f"Add new gpx tracks for user: {user}")

        path = Path(options.get("path"))
        path = path.expanduser()
        path = path.resolve()
        if not path.is_dir():
            self.stderr.write(f"ERROR: Given path '{path}' is not a existing directory!")
            sys.exit(4)

        self.stdout.write(f"Read directory: {path}")
        self.stdout.write("\n")

        new_tracks = 0
        for no, instance in enumerate(add_from_files(gpx_files_file_path=path, user=user), 1):
            self.stdout.write(self.style.SUCCESS("%i - Add new track: %s" % (no, instance)))
            new_tracks += 1

        self.stdout.write("\nimport done.\n")
        self.stdout.write(self.style.SUCCESS("Added %i new gpx tracks." % new_tracks))

        total_count = GpxModel.objects.filter(tracked_by=user).count()
        self.stdout.write("User %s has now %i tracks." % (user, total_count))
        self.stdout.write("\n")



"""
    created 25.08.2019 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import locale
import os
import sys

from django.core.management import BaseCommand as DjangoBaseCommand


class BaseCommand(DjangoBaseCommand):
    def execute(self, *args, **kwargs):
        """
        Just display a hint on UnicodeEncodeError

        e.g.:
        terminal encoding fallback to ASCII (e.g.: wrong server config in a ssh session)
        call gpx import command and the track contains non ASCII characters

         File ".../for_runners/management/commands/import_gpx.py", line 70, in handle
            self.stdout.write(self.style.SUCCESS("%i - Add new track: %s" % (no, instance)))
          File ".../python3.6/site-packages/django/core/management/base.py", line 145, in write
            self._out.write(style_func(msg))
        UnicodeEncodeError: 'ascii' codec can't encode character '\xdf' in position 42:...
        """
        self.stdout.write("\n")
        self.stdout.write("_" * 79)
        self.stdout.write(self.help)
        self.stdout.write("\n")

        try:
            return super().execute(*args, **kwargs)
        except UnicodeEncodeError as err:

            self.stderr.write("\n")
            self.stderr.write("*" * 79)
            self.stderr.write(f"UnicodeEncodeError: {err}")
            self.stderr.write("\n")
            self.stderr.write("Hint:")
            self.stderr.write(" - Maybe python output encoding falls back to ASCII ?")
            self.stderr.write(" - setup your locales or set LANG or PYTHONIOENCODING ;)")
            self.stderr.write("\n")
            self.stderr.write(f"locale: {repr(locale.getlocale())}")
            self.stderr.write(f"LANG: {os.environ.get('LANG')!r}")
            self.stderr.write(f"PYTHONIOENCODING: {os.environ.get('PYTHONIOENCODING')!r}")
            self.stderr.write(f"sys.stdout.encoding: {sys.stdout.encoding!r}")
            self.stderr.write("\n")
            self.stderr.write("*" * 79)
            self.stderr.write("\n")
            raise



"""
    created 12.06.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from pathlib import Path

from django.test import TestCase
# https://github.com/jedie/django-tools
from django_tools.unittest_utils.assertments import assert_pformat_equal
from django_tools.unittest_utils.user import TestUserMixin

# https://github.com/jedie/django-for-runners
from for_runners.services.gpx_create import add_from_file, add_from_files, add_gpx


BASE_PATH = Path(__file__).parent


class GpxTests(TestUserMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.user = self._get_user(usertype="normal")

    def assert_garmin_connect_1_gpx(self, instance):
        print(instance)

        self.assertEqual(repr(instance), "<GpxModel: 2018-02-21 Moers>")

        self.assertEqual(instance.points_no, 3)
        self.assertEqual(round(instance.length, 3), 4.727)
        self.assertEqual(instance.duration_s, 2)
        self.assertEqual(round(instance.pace, 3), 7.052)
        self.assertEqual(instance.heart_rate_avg, 125)

        self.assertEqual(instance.get_short_slug(), "2018-02-21-moers")

    def test_add_gpx(self):
        filepath = Path(BASE_PATH, "fixture_files/garmin_connect_1.gpx")
        with filepath.open("r") as f:
            gpx_content = f.read()

        instance = add_gpx(gpx_content=gpx_content, user=self.user)
        self.assert_garmin_connect_1_gpx(instance)

    def test_add_from_file(self):
        instance = add_from_file(
            gpx_file_file_path=Path(BASE_PATH, "fixture_files/garmin_connect_1.gpx"), user=self.user
        )
        self.assert_garmin_connect_1_gpx(instance)

    def test_add_from_files(self):
        instances = [
            str(instance)
            for instance in add_from_files(
                gpx_files_file_path=Path(BASE_PATH, "fixture_files"),
                user=self.user,
                skip_errors=True,
            )
        ]
        assert_pformat_equal(
            instances, ["2018-02-21 Moers", "2011-01-13 Berlin Tiergarten"]
        )



"""
    Helper to call django manage commands
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    used in
    for, e.g.:

    ~/DjangoForRunnersEnv/bin/python3 -m for_runners_project --help
"""
import os
import sys

from django.core.management import execute_from_command_line
from django_tools.exception_plus import print_exc_plus


def manage():
    os.environ["DJANGO_SETTINGS_MODULE"] = "for_runners_project.settings"
    try:
        execute_from_command_line(sys.argv)
    except SystemExit as err:
        sys.exit(err.code)
    except BaseException:
        print_exc_plus()
        sys.exit(-1)


if __name__ == "__main__":
    manage()



"""
    created 17.11.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import os
import sys
from pathlib import Path


class VirtualEnvPath:
    def __init__(self):
        # print("sys.executable:", sys.executable)
        # print("sys.prefix:", sys.prefix)
        # print("sys.real_prefix:", getattr(sys, "real_prefix", None))

        self.env_path = Path(sys.prefix)  # e.g.: /home/<username>/DjangoForRunnersEnv
        assert self.env_path.is_dir()

    def get_for_runners_exe(self):
        if sys.platform in ("win32", "cygwin"):
            for_runners_filename = "for_runners.exe"
        else:
            for_runners_filename = "for_runners"

        self.executable_path = Path(sys.executable)  # e.g.: /home/<username>/DjangoForRunnersEnv/bin/python3
        assert self.executable_path.is_file()

        # raise ValueError if self.env_path is not a subpath of self.executable_path
        self.executable_path.relative_to(self.env_path)

        for_runners_exe = Path(self.executable_path.parent, for_runners_filename)
        assert for_runners_exe.is_file(), "for_runner executeable not found here: '%s'" % for_runners_exe
        assert os.access(str(for_runners_exe), os.X_OK), "File not executeable: '%s'" % for_runners_exe

        return for_runners_exe


def get_venv_path():
    """
    :return: VirtualEnv root dir, e.g.: /home/<username>/DjangoForRunnersEnv
    """
    return VirtualEnvPath().env_path



"""
    created 28.06.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


from django import forms
from django.conf import settings

# https://github.com/jedie/django-for-runners
from for_runners.management.commands.base import BaseCommand
from for_runners.models import DistanceModel, GpxModel


class DistanceForm(forms.ModelForm):
    class Meta:
        model = DistanceModel
        fields = ['distance_km']


class Command(BaseCommand):
    help = "Create some base data"

    def handle(self, *args, **options):

        updated_needed = False

        for distance_km in settings.BASE_IDEAL_TRACK_LENGTHS:
            form = DistanceForm(data={'distance_km': distance_km})
            if not form.is_valid():
                raise AssertionError(
                    f'settings.BASE_IDEAL_TRACK_LENGTHS {distance_km!r} not valid: {form.errors["distance_km"]}'
                )

            distance_km = form.cleaned_data['distance_km']
            obj, created = DistanceModel.objects.get_or_create(distance_km=distance_km)
            if created:
                updated_needed = True
                self.stdout.write(f"Create: {repr(obj)}")
            else:
                self.stdout.write(f"{repr(obj)} already exists, ok.")

        if not updated_needed:
            print("No track updated needed, ok.")
        else:
            for track in GpxModel.objects.all():
                track.save()
                print(".", end="", flush=True)
            print()



"""
    created 28.06.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
from django.db.models.signals import post_save, pre_save

# https://github.com/jedie/django-for-runners
from for_runners.models.discipline import DisciplineModel  # noqa
from for_runners.models.distance import DistanceModel  # noqa
from for_runners.models.event import CostModel, EventLinkModel, EventModel, ParticipationModel  # noqa
from for_runners.models.gpx import GpxModel  # noqa
from for_runners.signal_handlers.gpx import gpx_post_save_handler, gpx_pre_save_handler  # noqa


pre_save.connect(receiver=gpx_pre_save_handler, sender=GpxModel)
post_save.connect(receiver=gpx_post_save_handler, sender=GpxModel)



#!/usr/bin/env python3

import os
import sys
import traceback
from pathlib import Path
from threading import Timer

from django.contrib.auth import get_user_model
from django.contrib.staticfiles.management.commands.runserver import Command as RunServerCommand
from django.core.management import call_command

# https://github.com/jedie/django-for-runners
from for_runners.models import DistanceModel


DONT_OPEN_BROWSER_ENV_KEY = "DONT_OPEN_BROWSER"


class Command(RunServerCommand):
    """
    Expand django.contrib.staticfiles runserver
    """

    help = "Run Django-ForRunners with django developer server"

    def verbose_call(self, command, *args, **kwargs):
        self.stderr.write("_" * 79)
        self.stdout.write(f"Call {command!r} with: {args!r} {kwargs!r}")
        call_command(command, *args, **kwargs)

    def handle(self, *args, **options):

        if "RUN_MAIN" not in os.environ:
            # RUN_MAIN added by auto reloader, see: django/utils/autoreload.py

            # normally we can just do the following here::
            #   call_command("makemigrations")
            #   call_command("migrate")
            #
            # But this work until fix of: https://bitbucket.org/kbr/autotask/pull-requests/3/
            #
            # work-a-round: We call it in the shell script

            # django.contrib.staticfiles.management.commands.collectstatic.Command
            if sys.platform in ("win32", "cygwin"):
                # fix: CommandError: symbolic link privilege not held
                link = False
            else:
                link = True
            self.verbose_call("collectstatic", interactive=False, link=link)

            User = get_user_model()
            qs = User.objects.filter(is_active=True, is_superuser=True)
            if qs.count() == 0:
                self.verbose_call("createsuperuser")

            distance_model_count = DistanceModel.objects.all().count()
            if distance_model_count == 0:
                # call: for_runners.management.commands.fill_basedata.Command
                self.verbose_call("fill_basedata")

            if DONT_OPEN_BROWSER_ENV_KEY not in os.environ:

                def open_webbrowser():
                    os.environ[DONT_OPEN_BROWSER_ENV_KEY] = "yes"  # open only one time ;)

                    self.stderr.write("_" * 79)
                    self.stderr.write("Start web server...")

                    import webbrowser

                    addr = f"[{self.addr}]" if self._raw_ipv6 else self.addr
                    uri = f"{self.protocol}://{addr}:{self.port}/"
                    print(f"\nStart browser with: {uri!r}\n")
                    webbrowser.open_new_tab(uri)

                Timer(1, open_webbrowser).start()

        options["insecure_serving"] = True
        super().handle(*args, **options)

    def run(self, **options):

        if sys.platform in ("win32", "cygwin"):
            # Bugfix for:
            #   can't open file 'C:\Program Files\Django-ForRunners\Scripts\for_runners':
            #   [Errno 2] No such file or directory
            executable = Path(Path(sys.argv[0]).parent, "for_runners-script.py")
            print(f"Patch executeable to: {executable}")
            assert executable.is_file(), f"Executeable not found here: {executable}"
            sys.argv[0] = str(executable)

        try:
            super().run(**options)
        except Exception:
            traceback.print_last()
            raise



"""
    Test API connection to metaweather.com via for_runners.weather

    https://www.metaweather.com/api/

    created 21.06.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import datetime

from for_runners.tests.base import BaseTestCase
from for_runners.weather import NoWeatherData, meta_weather_com


class WeatherTest(BaseTestCase):
    """
    TODO: Mock request!
    """

    def test(self):
        lat, lon = (51.4109, 6.7828)  # Duisburg -> WOEID: 648820 (Essen, city)
        date = datetime.datetime(year=2018, month=6, day=20, hour=20, minute=30)

        # Essen City on 21.06.2018
        # https://www.metaweather.com/de/648820/2018/6/20/
        temperature, weather_state = meta_weather_com.coordinates2weather(
            lat, lon, date=date, max_seconds=12 * 60 * 60
        )

        self.assert_equal_rounded(temperature, 25.41, decimal_places=2)
        self.assertEqual(weather_state, "Light Cloud/Showers")

    def test_small_max_seconds(self):
        lat, lon = (51.4109, 6.7828)  # Duisburg -> WOEID: 648820 (Essen, city)
        date = datetime.datetime(year=2018, month=6, day=20, hour=20, minute=30)

        # Essen City on 21.06.2018
        # https://www.metaweather.com/de/648820/2018/6/20/
        temperature, weather_state = meta_weather_com.coordinates2weather(lat, lon, date=date, max_seconds=0.1)

        self.assert_equal_rounded(temperature, 27.94, decimal_places=2)
        self.assertEqual(weather_state, "Light Cloud")

    def test_no_json_data(self):
        """
        Request weather for Essen City on 11.02.2017 will raise into 500 - Server error:
            https://www.metaweather.com/de/648820/2017/2/11/

        The API Request will return []:
            https://www.metaweather.com/api/location/648820/2017/2/11/

        MetaWeatherCom().location_day() will raise NoWeatherData
        """
        lat, lon = (51.4109, 6.7828)  # Duisburg -> WOEID: 648820 (Essen, city)
        date = datetime.datetime(year=2017, month=2, day=10, hour=12, minute=00)

        with self.assertRaises(NoWeatherData):
            temperature, weather_state = meta_weather_com.coordinates2weather(lat, lon, date=date)



"""
    created 30.05.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import logging

from django.contrib import admin

# https://github.com/jedie/django-for-runners
from for_runners.models import DisciplineModel


log = logging.getLogger(__name__)


@admin.register(DisciplineModel)
class DisciplineModelAdmin(admin.ModelAdmin):
    pass



"""
    Django settings used while pytest run
"""

from for_runners_project.settings import *  # noqa


# https://bitbucket.org/kbr/autotask
#
# avoid:
#
# Error in atexit._run_exitfuncs:
# Traceback (most recent call last):
#   File ".../lib/python3.6/site-packages/autotask/apps.py", line 27, in __call__
#     self.delete_periodic_tasks()
#   File ".../lib/python3.6/site-packages/autotask/apps.py", line 37, in delete_periodic_tasks
#     if qs.count():
#   ...
# Failed: Database access not allowed, use the "django_db" mark, or the
# "db" or "transactional_db" fixtures to enable it.
AUTOTASK_IS_ACTIVE = False


# Skip download map via geotiler in for_runners.gpx_tools.gpxpy2map.generate_map
MAP_DOWNLOAD = False



import sys

from for_runners import __version__


if __name__ == "for_runners_project":
    #
    # This will be called before the click cli
    #
    if "--version" not in sys.argv:
        print(f"Django-ForRunners v{__version__}")

    if len(sys.argv) == 1:
        # FIXME: How can be a "default" action set in click?
        sys.argv.append("run-server")



"""
Based on http://docs.gunicorn.org/en/stable/custom.html
"""

import multiprocessing

import gunicorn.app.base
from django.core.wsgi import get_wsgi_application


class GunicornApplication(gunicorn.app.base.BaseApplication):
    def __init__(self, application, **options):
        self.options = options
        self.application = application
        super().__init__()

    def load_config(self):
        for key, value in self.options.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        return self.application


def get_gunicorn_application(address="127.0.0.1:8000"):
    print(f"Start gunicorn on: {address!r}")
    wsgi_application = get_wsgi_application()
    number_of_workers = (multiprocessing.cpu_count() * 2) + 1
    return GunicornApplication(wsgi_application, bind=address, workers=number_of_workers)



from pathlib import Path

from django.core.files.uploadedfile import UploadedFile
from django_tools.unittest_utils.unittest_base import BaseTestCase
from django_tools.unittest_utils.user import TestUserMixin

from for_runners import __version__
from for_runners.models import GpxModel


BASE_PATH = Path(__file__).parent


class ForRunnerAdminTests(TestUserMixin, BaseTestCase):
    """
        Special for runners admin tests
    """

    def test_staff_upload(self):
        self.login(usertype="superuser")

        gpx_file_path1 = Path(BASE_PATH, "fixture_files/garmin_connect_1.gpx")
        gpx_file_path2 = Path(BASE_PATH, "fixture_files/no_track_points.gpx")

        with gpx_file_path1.open("rb") as file1, gpx_file_path2.open("rb") as file2:

            response = self.client.post(
                "/en/admin/for_runners/gpxmodel/upload/",
                data={
                    "gpx_files": [
                        UploadedFile(file=file1, name=gpx_file_path1.name, content_type="application/gpx+xml"),
                        UploadedFile(file=file2, name=gpx_file_path2.name, content_type="application/gpx+xml"),
                    ]
                },
                HTTP_ACCEPT_LANGUAGE="en",
            )
            # debug_response(response)

        tracks = GpxModel.objects.all()
        self.assertEqual(tracks.count(), 1)
        new_track = tracks[0]

        self.assertRedirects(
            response,
            expected_url=f"/en/admin/for_runners/gpxmodel/{new_track.pk:d}/change/",
            status_code=302,
            target_status_code=200,
            fetch_redirect_response=False,
        )

        response = self.client.get("/en/admin/for_runners/gpxmodel/", HTTP_ACCEPT_LANGUAGE="en")
        # print(repr(self.get_messages(response)))
        self.assertResponse(
            response,
            must_contain=(
                f"<title>Select GPX Track to change | Django-ForRunners v{__version__}</title>",
                # "Process garmin_connect_1.gpx...",
                # "Created: 2018-02-21 14:30:50 Moers",
                #
                # "Process no_track_points.gpx...",
                # "Error process GPX data: Can't get first track",
                '<td class="field-human_pace">7:03 min/km</td>',
                '<p class="paginator">1 GPX Track</p>',
            ),
            messages=[
                "Process garmin_connect_1.gpx...",
                "Created: 2018-02-21 Moers",
                "Process no_track_points.gpx...",
                "Error process GPX data: Can't get first track",
            ],
            must_not_contain=("error", "traceback"),
            template_name="admin/for_runners/gpxmodel/change_list.html",
            html=True,
        )

    def test_add_view_redirect_to_upload(self):
        self.login(usertype="superuser")
        response = self.client.get("/en/admin/for_runners/gpxmodel/add/", HTTP_ACCEPT_LANGUAGE="en")
        self.assertRedirects(
            response,
            expected_url="/en/admin/for_runners/gpxmodel/upload/",
            status_code=302,
            target_status_code=200,
            fetch_redirect_response=True,
        )



from django.conf import settings
from django.core.management import call_command
from django.test import TestCase

# https://github.com/jedie/django-for-runners
from for_runners.management.commands.fill_basedata import Command as FillBasedataCommand
from for_runners.models import DistanceModel


class FillBasedataCommandTestCase(TestCase):
    def test_import_no_username_given(self):
        assert len(settings.BASE_IDEAL_TRACK_LENGTHS) > 0
        assert DistanceModel.objects.count() == 0

        call_command(FillBasedataCommand())

        assert DistanceModel.objects.count() == len(settings.BASE_IDEAL_TRACK_LENGTHS)



import sys
from pathlib import Path

import click

# https://github.com/jedie/django-for-runners
import for_runners
from for_runners_project.utils.venv import VirtualEnvPath


SVG_LOGO = "static/Django-ForRunners.svg"  # .../for_runners/static/Django-ForRunners.svg

XDG_OPEN_FILENAME_NORMAL = "Django-ForRunners"
XDG_OPEN_FILENAME_DEVELOP = "Django-ForRunners develop"
XDG_OPEN_TEMPLATE = """#!/usr/bin/env xdg-open

[Desktop Entry]
Version=1.0
Type=Application
Terminal=false
Icon={svg_logo_path}
Exec=x-terminal-emulator -e "{for_runners_exe} {command}"
Name={name}
"""

# gnome-terminal -x bash -c '/usr/bin/cal && bash'


def get_for_runners_app_path():
    for_runners_app_path = Path(for_runners.__file__).parent
    return for_runners_app_path


def get_svg_logo_path():
    for_runners_app_path = get_for_runners_app_path()
    svg_logo_path = Path(for_runners_app_path, SVG_LOGO)
    assert svg_logo_path.is_file(), f"Logo not found here: {svg_logo_path}"
    return svg_logo_path


def create_linux_xdg_open_file(file_name, for_runners_exe, command, env_path):
    desktop_file_path = Path(env_path, f"{file_name}.desktop")
    print(f"Create linux xdg-open starter here: {desktop_file_path}")

    svg_logo_path = get_svg_logo_path()
    content = XDG_OPEN_TEMPLATE.format(
        name=file_name, svg_logo_path=svg_logo_path, for_runners_exe=for_runners_exe, command=command
    )
    with desktop_file_path.open("w") as f:
        f.write(content)
    desktop_file_path.chmod(0o777)


def create_starter():
    click.echo("Create stater")

    venv_path = VirtualEnvPath()
    env_path = venv_path.env_path
    print(f"Create starter in: {env_path}")

    for_runners_exe = venv_path.get_for_runners_exe()

    if sys.platform in ("win32", "cygwin"):
        raise NotImplementedError("TODO: Create starter under Windows!")
    else:
        create_linux_xdg_open_file(XDG_OPEN_FILENAME_NORMAL, for_runners_exe, "run_gunicorn", env_path)
        create_linux_xdg_open_file(XDG_OPEN_FILENAME_DEVELOP, for_runners_exe, "run_dev_server", env_path)



from django.conf.urls import include, static, url
from django.conf.urls.i18n import i18n_patterns
from django.contrib import admin
from django.urls import path
from django.views.generic import RedirectView

# https://github.com/jedie/django-for-runners
from for_runners_project import settings


admin.autodiscover()

urlpatterns = i18n_patterns(
    path("admin/", admin.site.urls),
    # until there is not real CMS pages: redirect to the interesting admin page:
    url(r"^$", RedirectView.as_view(url="/admin/for_runners/gpxmodel/")),
)

if settings.DEBUG:
    urlpatterns += static.static(settings.STATIC_URL, document_root=settings.STATIC_ROOT)
    urlpatterns += static.static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)

    import debug_toolbar

    urlpatterns = [url(r"^__debug__/", include(debug_toolbar.urls))] + urlpatterns



"""
    created 30.05.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import collections
import logging
import math
import statistics
from pprint import pprint

from django import forms
from django.conf.urls import url
from django.contrib import admin, messages
from django.contrib.admin.views.main import ChangeList
from django.db import IntegrityError, NotSupportedError, models
from django.db.models import Avg, Max, Min
from django.http import HttpResponseRedirect
from django.template.loader import render_to_string
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _
from django.views import generic
# https://github.com/jedie/django-tools
from django_tools.decorators import display_admin_error
from import_export.admin import ExportMixin

# https://github.com/jedie/django-for-runners
from for_runners import constants
from for_runners.admin.gpx_import_export import GpxModelResource
from for_runners.admin.utils import BaseChangelistView, BaseFormChangelistView
from for_runners.exceptions import GpxDataError
from for_runners.forms import INITIAL_DISTANCE, DistanceStatisticsForm, UploadGpxFileForm
from for_runners.gpx import add_extension_data, get_2d_coordinate_list, iter_distance, iter_points
from for_runners.models import GpxModel

from for_runners.services.gpx_svg_generator import generate_svg


log = logging.getLogger(__name__)

STATISTICS_CHOICES = (
    (constants.DISPLAY_DISTANCE_PACE_KEY, _("Distance/Pace")),
    (constants.DISPLAY_PACE_DURATION, _("Pace/Duration")),
    (constants.DISPLAY_GPX_INFO, _("GPX info")),
    (constants.DISPLAY_GPX_METADATA, _("GPX metadata")),
)
assert len(dict(STATISTICS_CHOICES)) == len(STATISTICS_CHOICES), "Double keys?!?"


class UploadGpxFileView(generic.FormView):
    template_name = "for_runners/upload_gpx_file.html"
    form_class = UploadGpxFileForm
    success_url = "../"  # FIXME

    def post(self, request, *args, **kwargs):
        user = request.user
        form_class = self.get_form_class()
        form = self.get_form(form_class)
        files = request.FILES.getlist("gpx_files")
        if form.is_valid():
            log.debug("files: %r", files)
            for f in files:
                messages.info(request, f"Process {f.name}...")

                content = f.file.read()
                log.debug("raw content......: %s", repr(content)[:100])
                content = content.decode("utf-8")
                log.debug("decoded content..: %s", repr(content)[:100])

                try:
                    try:
                        gpx = GpxModel.objects.create(gpx=content, tracked_by=user)
                    except IntegrityError as err:
                        # catch: UNIQUE constraint failed
                        # give a better error message
                        messages.error(request, f"Error process GPX data: {err}")
                        continue
                except GpxDataError as err:
                    messages.error(request, f"Error process GPX data: {err}")
                else:
                    messages.success(request, f"Created: {gpx}")

                    # redirect to change view:
                    self.success_url = gpx.get_admin_change_url()

            return self.form_valid(form)
        else:
            return self.form_invalid(form)


class DistancePaceStatisticsView(BaseChangelistView):
    template_name = "for_runners/distance_pace_statistics.html"

    def get_context_data(self, **kwargs):
        qs = self.change_list.queryset  # get the filteres queryset form GpxModelChangeList
        qs = qs.order_by("length")
        context = {
            "tracks": qs,
            "track_count": qs.count(),
            "title": _("Distance/Pace Statistics"),
            "user": self.request.user,
            "opts": GpxModel._meta,
        }
        return context


class DistanceStatisticsView(BaseFormChangelistView):
    template_name = "for_runners/distance_statistics.html"
    form_class = DistanceStatisticsForm

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        form = context["form"]
        if form.is_valid():
            distance = form.cleaned_data["distance"]
        else:
            distance = INITIAL_DISTANCE

        distance_m = distance * 1000

        qs = self.change_list.queryset  # get the filteres queryset form GpxModelChangeList
        qs = qs.order_by("length")

        length_statistics = qs.aggregate(Min("length"), Avg("length"), Max("length"))

        min_length = length_statistics["length__min"]
        max_length = length_statistics["length__max"]

        current_distance_from = math.floor(min_length / 1000) * 1000
        current_distance_to = current_distance_from + distance_m

        group_data = collections.defaultdict(list)
        for track in qs:
            length = track.length
            # print("%.1fkm" % round(length/1000,1))
            if length > current_distance_to:
                while True:
                    current_distance_from += distance_m
                    current_distance_to += distance_m
                    if length > current_distance_to:
                        group_data[(current_distance_from, current_distance_to)] = []
                    else:
                        break

            group_data[(current_distance_from, current_distance_to)].append(track)

        print("group_data:")
        pprint(group_data)

        track_data = []
        total_tracks = 0
        for distances, tracks in sorted(group_data.items()):
            track_count = len(tracks)
            total_tracks += track_count
            distance_from, distance_to = distances

            if tracks:
                paces = [track.pace for track in tracks]
                min_paces = f"{min(paces):.2f}"
                avg_paces = f"{statistics.median(paces):.2f}"
                max_paces = f"{max(paces):.2f}"
            else:
                min_paces = "null"
                avg_paces = "null"
                max_paces = "null"

            track_data.append(
                (
                    round(distance_from / 1000, 1),
                    round(distance_to / 1000, 1),
                    track_count,
                    min_paces,
                    avg_paces,
                    max_paces,
                )
            )
        print("total track counts:", total_tracks)
        pprint(track_data)

        context.update(
            {
                "tracks": qs,
                "track_count": total_tracks,
                "min_length_km": round(min_length / 1000),
                "avg_length_km": round(length_statistics["length__avg"] / 1000),
                "max_length_km": round(max_length / 1000),
                "track_data": track_data,
                "title": _("Distance Statistics"),
                "user": self.request.user,
                "opts": GpxModel._meta,
            }
        )
        # pprint(context)
        return context


class GpxInfoView(BaseChangelistView):
    template_name = "for_runners/gpx_info.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(
            {
                "tracks": self.change_list.queryset,  # get the filteres queryset form GpxModelChangeList,
                "title": _("GPX Infomation"),
                "user": self.request.user,
                "opts": GpxModel._meta,
            }
        )
        return context


class GpxMetadataView(BaseChangelistView):
    template_name = "for_runners/gpx_metadata.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(
            {
                "tracks": self.change_list.queryset,  # get the filteres queryset form GpxModelChangeList
                "title": _("GPX Metadata"),
                "user": self.request.user,
                "opts": GpxModel._meta,
            }
        )
        return context


class PrintMiniView(generic.TemplateView):
    """
    Generate a page to print a small overview.
    """

    template_name = "admin/for_runners/gpxmodel/print_mini.html"

    def get(self, request):
        ids = request.GET["ids"].split(",")
        ids = [int(pk) for pk in ids]
        self.instances = GpxModel.objects.filter(pk__in=ids)
        return super().get(request)

    def get_context_data(self, **kwargs):
        gpx_tracks = []
        for gpx_track in self.instances:
            if not gpx_track.track_svg.name:
                log.critical("GPX Track has no svg file: %s", gpx_track)
                generate_svg(gpx_track, force=False)

            gpx_tracks.append(gpx_track)
        context = {"gpx_tracks": gpx_tracks, "back_url": reverse("admin:for_runners_gpxmodel_changelist")}
        return context


class CalculateValuesView(generic.View):
    def get(self, request, object_id):
        instance = GpxModel.objects.get(pk=object_id)
        instance.calculate_values()
        instance.save()
        messages.success(request, "Values are calculated from GPX data")
        return HttpResponseRedirect("../")


class StatisticsListFilter(admin.SimpleListFilter):
    title = _("statistics")
    template = "admin/for_runners/gpxmodel/filter.html"

    # Parameter for the filter that will be used in the URL query.
    parameter_name = constants.STATISTICS_PARAMETER_NAME

    def lookups(self, request, model_admin):
        return STATISTICS_CHOICES

    def queryset(self, request, queryset):
        return queryset


class GpxModelChangeList(ChangeList):
    def __init__(self, *args, **kwargs):
        self.startistics_mapping = {
            constants.DISPLAY_DISTANCE_PACE_KEY: DistanceStatisticsView,
            constants.DISPLAY_PACE_DURATION: DistancePaceStatisticsView,
            constants.DISPLAY_GPX_INFO: GpxInfoView,
            constants.DISPLAY_GPX_METADATA: GpxMetadataView,
        }

        # work-a-round for:
        #   __init__() missing 1 required positional argument: 'sortable_by'
        # while using export view
        # kwargs["sortable_by"] = None

        super().__init__(*args, **kwargs)

    def get_results(self, request):
        super().get_results(request)

        self.statistics = ""

        if constants.STATISTICS_PARAMETER_NAME in request.GET:
            if self.result_count == 0:
                log.debug("No tracks: no statistics.")
                return

            key = request.GET[constants.STATISTICS_PARAMETER_NAME]
            try:
                ViewClass = self.startistics_mapping[key]
            except KeyError as err:
                log.error("statistic view unknown: %s", err)
            else:
                view = ViewClass.as_view()
                response = view(request, self)
                assert isinstance(response, TemplateResponse), f"Method {view} didn't return a TemplateResponse!"
                self.statistics = response.rendered_content


class HasNetDurationFilter(admin.SimpleListFilter):
    title = _("has net duration")
    parameter_name = "net_duration"

    def lookups(self, request, model_admin):
        return (("y", _("yes")), ("n", _("no")))

    def queryset(self, request, queryset):
        if self.value() == "y":
            return queryset.exclude(participation__duration__isnull=True)
        if self.value() == "n":
            return queryset.filter(participation__duration__isnull=True)


class HasEventPartricipationFilter(admin.SimpleListFilter):
    title = _("has event participation")
    parameter_name = "participation"

    def lookups(self, request, model_admin):
        return (("y", _("yes")), ("n", _("no")))

    def queryset(self, request, queryset):
        if self.value() == "y":
            return queryset.exclude(participation__isnull=True)
        if self.value() == "n":
            return queryset.filter(participation__isnull=True)


@admin.register(GpxModel)
class GpxModelAdmin(ExportMixin, admin.ModelAdmin):
    actions = ["print_mini"]
    # change_list_template = 'admin/import_export/change_list_export.html'
    change_list_template = "admin/for_runners/gpxmodel/change_list.html"
    resource_class = GpxModelResource

    def print_mini(self, request, queryset):
        url = reverse("admin:print-mini")  # for_runners.admin.gpx.PrintMiniView
        selected = request.POST.getlist(admin.ACTION_CHECKBOX_NAME)
        return HttpResponseRedirect(f"{url}?ids={','.join(selected)}")

    print_mini.short_description = _("Generate a page to print a small overview.")

    def add_view(self, request, form_url="", extra_context=None):
        # redirect the defaul add view to upload form view:
        return HttpResponseRedirect(reverse("admin:upload-gpx-file"))

    @display_admin_error
    def leaflet_map_html(self, obj):
        gpxpy_instance = obj.get_gpxpy_instance()

        lat_list, lon_list = get_2d_coordinate_list(gpxpy_instance)
        coordinates = zip(lat_list, lon_list)

        km_gpx_points = iter_distance(gpxpy_instance, distance=1000)

        context = {
            "short_start_address": obj.short_start_address,
            "start_time": obj.start_time,
            "start_latitude": obj.start_latitude,
            "start_longitude": obj.start_longitude,
            "short_finish_address": obj.short_finish_address,
            "finish_time": obj.finish_time,
            "finish_latitude": obj.finish_latitude,
            "finish_longitude": obj.finish_longitude,
            "coordinates": coordinates,
            "km_gpx_points": km_gpx_points,
        }
        return render_to_string(template_name="admin/for_runners/gpxmodel/leaflet_map.html", context=context)

    leaflet_map_html.short_description = _("Route")

    @display_admin_error
    def dygraphs_html(self, obj):
        """
        Use dygraphs array format:
            http://dygraphs.com/data.html#array
        """
        gpxpy_instance = obj.get_gpxpy_instance()

        has_hr = None
        has_cad = None

        elevation_label = _("Elevation")

        labels = [_("Date"), elevation_label]
        columns = []

        time2coordinates = {}

        for point in iter_points(gpxpy_instance):
            add_extension_data(point)

            timestamp = point.time.timestamp() * 1000

            time2coordinates[timestamp] = (point.latitude, point.longitude)

            row = ["new Date(%i)" % timestamp, point.elevation]

            if has_hr is None or has_hr:
                try:
                    row.append(point.extension_data["hr"])
                except KeyError:
                    has_hr = False
                else:
                    if has_hr is None:
                        has_hr = True
                        labels.append(_("heart rate"))

            if has_cad is None or has_cad:
                try:
                    row.append(point.extension_data["cad"])
                except KeyError:
                    has_cad = False
                else:
                    if has_cad is None:
                        has_cad = True
                        labels.append(_("cadence"))

            columns.append(",".join([str(i) for i in row]))

        km_points = []
        for point, distance_m, distance_km in iter_distance(gpxpy_instance, distance=1000):
            km_points.append(
                {"x": "%i" % (point.time.timestamp() * 1000), "distance_m": distance_m, "distance_km": distance_km}
            )

        context = {
            "instance": obj,
            "labels": labels,
            "columns": columns,
            "elevation_label": elevation_label,
            "km_points": km_points,
            "time2coordinates": time2coordinates,
        }
        return render_to_string(template_name="admin/for_runners/gpxmodel/dygraphs.html", context=context)

    dygraphs_html.short_description = _("Graphs")

    def participation_links(self, obj):
        if not obj.participation:
            return "-"

        event = obj.participation.event
        gpx_tracks = GpxModel.objects.all().filter(participation__event=event).exclude(pk=obj.pk)

        context = {"event": event, "gpx_tracks": gpx_tracks}
        return render_to_string(template_name="admin/for_runners/gpxmodel/participation_links.html", context=context)

    participation_links.short_description = _("Links")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.list_display = [
            "tracked_by",
            "svg_tag",
            "overview",
            "start_time",
            "human_length_html",
            "human_duration_html",
            "human_pace",
            "heart_rate_avg",
            "human_weather",
            "uphill",
            "downhill",
            "min_elevation",
            "max_elevation",
        ]
        self.list_filter = [
            StatisticsListFilter,
            HasNetDurationFilter,
            HasEventPartricipationFilter,
            "tracked_by",
            "start_time",
            "ideal_distance",
            "creator",
        ]

    search_fields = ("full_start_address", "full_finish_address", "creator")
    date_hierarchy = "start_time"
    list_per_page = 50
    list_display_links = ("svg_tag", "overview")
    readonly_fields = (
        "leaflet_map_html",
        "dygraphs_html",
        "participation_links",
        "svg_tag_big",
        "svg_tag",
        "start_time",
        "start_latitude",
        "start_longitude",
        "finish_time",
        "finish_latitude",
        "finish_longitude",
        "start_coordinate_html",
        "finish_coordinate_html",
        "heart_rate_min",
        "heart_rate_avg",
        "heart_rate_max",
        "human_length_html",
        "human_duration_html",
        "human_pace",
    )

    fieldsets = (
        (_("Map / Graphs"), {"fields": ("leaflet_map_html", "dygraphs_html")}),
        (_("Event"), {"fields": ("participation", "participation_links")}),
        (
            _("Start"),
            {
                "fields": (
                    ("start_time", "start_temperature", "start_weather_state"),
                    "short_start_address",
                    "full_start_address",
                    "start_coordinate_html",
                    ("start_latitude", "start_longitude"),
                )
            },
        ),
        (
            _("Finish"),
            {
                "fields": (
                    ("finish_time", "finish_temperature", "finish_weather_state"),
                    "short_finish_address",
                    "full_finish_address",
                    "finish_coordinate_html",
                    ("finish_latitude", "finish_longitude"),
                )
            },
        ),
        (_("GPX data"), {"classes": ("collapse",), "fields": (("gpx", "points_no"), ("track_svg", "svg_tag_big"))}),
        (
            _("Values"),
            {
                "fields": (
                    ("human_length_html", "ideal_distance"),
                    ("human_duration_html", "human_pace"),
                    ("heart_rate_min", "heart_rate_avg", "heart_rate_max"),
                    ("uphill", "downhill"),
                    ("min_elevation", "max_elevation"),
                )
            },
        ),
    )
    # FIXME: Made this in CSS ;)
    formfield_overrides = {models.CharField: {"widget": forms.TextInput(attrs={"style": "width:70%"})}}

    def overview(self, obj):
        parts = []
        if obj.participation:
            event = obj.participation.event
            parts.append(f"<strong>{event}</strong>")
        parts.append(obj.start_end_address())
        html = "<br>".join(parts)
        return mark_safe(html)

    overview.short_description = _("Event")

    def get_urls(self):
        urls = super().get_urls()
        info = self.model._meta.app_label, self.model._meta.model_name
        urls = [
            url(r"^upload/$", self.admin_site.admin_view(UploadGpxFileView.as_view()), name="upload-gpx-file"),
            url(
                r"^distance-statistics/$",
                self.admin_site.admin_view(DistanceStatisticsView.as_view()),
                name="distance-statistics",
            ),
            url(
                r"^distance-pace-statistics/$",
                self.admin_site.admin_view(DistancePaceStatisticsView.as_view()),
                name="distance-pace-statistics",
            ),
            url(r"^print_mini/$", self.admin_site.admin_view(PrintMiniView.as_view()), name="print-mini"),
            url(
                r"^(.+)/calculate_values/$",
                self.admin_site.admin_view(CalculateValuesView.as_view()),
                name="%s_%s_calculate-values" % info,
            ),
        ] + urls
        return urls

    @cached_property
    def user_count(self):
        qs = GpxModel.objects.all().only("tracked_by").order_by("tracked_by")

        try:
            user_count = qs.distinct("tracked_by").count()
        except (NotImplementedError, NotSupportedError):
            # e.g.: sqlite has no distinct :(
            qs = qs.values_list("tracked_by__id", flat=True)
            user_count = len(set(qs))

        return user_count

    def get_list_display(self, request):
        list_display = super().get_list_display(request).copy()

        if self.user_count <= 1 and "tracked_by" in list_display:
            list_display.remove("tracked_by")

        return list_display

    def get_list_filter(self, request):
        list_filter = super().get_list_filter(request).copy()

        if self.user_count <= 1 and "tracked_by" in list_filter:
            list_filter.remove("tracked_by")

        return list_filter

    def get_changelist(self, request, **kwargs):
        """
        Returns the ChangeList class for use on the changelist page.
        """
        return GpxModelChangeList

    # def changelist_view(self, request, extra_context=None):
    #     if extra_context is None:
    #         extra_context = {}
    #

    #
    #     return super(GpxModelAdmin, self).changelist_view(request, extra_context)



# Generated by Django 2.0.7 on 2018-07-04 23:08

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import for_runners.models.gpx


class Migration(migrations.Migration):

    initial = True

    dependencies = [migrations.swappable_dependency(settings.AUTH_USER_MODEL)]

    operations = [
        migrations.CreateModel(
            name="DisciplineModel",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(help_text="Sport discipline", max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name="DistanceModel",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "distance_km",
                    models.DecimalField(
                        decimal_places=4, help_text="The ideal track length in kilometer.", max_digits=7, unique=True
                    ),
                ),
                (
                    "variance",
                    models.PositiveSmallIntegerField(
                        default=5, help_text="Maximum (+/-) deviation in percent to match this distance."
                    ),
                ),
                ("min_distance_m", models.PositiveIntegerField(blank=True, editable=False, null=True)),
                ("max_distance_m", models.PositiveIntegerField(blank=True, editable=False, null=True)),
            ],
            options={"verbose_name": "Distance", "verbose_name_plural": "Distances", "ordering": ("-distance_km",)},
        ),
        migrations.CreateModel(
            name="EventLinkModel",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "createtime",
                    models.DateTimeField(default=django.utils.timezone.now, editable=False, help_text="Create time"),
                ),
                (
                    "lastupdatetime",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False, help_text="Time of the last change."
                    ),
                ),
                ("url", models.URLField(help_text="Link URL")),
                (
                    "text",
                    models.CharField(
                        blank=True,
                        help_text="Link text (leave empty to generate it from url)",
                        max_length=127,
                        null=True,
                    ),
                ),
                (
                    "title",
                    models.CharField(
                        blank=True,
                        help_text="Link title (leave empty to generate it from url)",
                        max_length=127,
                        null=True,
                    ),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="EventModel",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "createtime",
                    models.DateTimeField(default=django.utils.timezone.now, editable=False, help_text="Create time"),
                ),
                (
                    "lastupdatetime",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False, help_text="Time of the last change."
                    ),
                ),
                ("no", models.PositiveIntegerField(blank=True, help_text="Sequential number of the event", null=True)),
                ("name", models.CharField(help_text="Name of the event", max_length=255)),
                ("start_date", models.DateField(blank=True, help_text="Start date of the run", null=True)),
                (
                    "createby",
                    models.ForeignKey(
                        blank=True,
                        editable=False,
                        help_text="User how create this entry.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="eventmodel_createby",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "discipline",
                    models.ForeignKey(
                        null=True, on_delete=django.db.models.deletion.SET_NULL, to="for_runners.DisciplineModel"
                    ),
                ),
                (
                    "lastupdateby",
                    models.ForeignKey(
                        blank=True,
                        editable=False,
                        help_text="User as last edit this entry.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="eventmodel_lastupdateby",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={"verbose_name": "Event", "verbose_name_plural": "Events", "ordering": ("-start_date", "-pk")},
        ),
        migrations.CreateModel(
            name="GpxModel",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "createtime",
                    models.DateTimeField(default=django.utils.timezone.now, editable=False, help_text="Create time"),
                ),
                (
                    "lastupdatetime",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False, help_text="Time of the last change."
                    ),
                ),
                ("gpx", models.TextField(help_text="The raw gpx file content")),
                (
                    "creator",
                    models.CharField(
                        blank=True, help_text="Used device to create this track", max_length=511, null=True
                    ),
                ),
                (
                    "track_svg",
                    models.FileField(
                        blank=True,
                        null=True,
                        upload_to=for_runners.models.gpx.svg_upload_path,
                        verbose_name="Track SVG",
                    ),
                ),
                (
                    "start_time",
                    models.DateTimeField(editable=False, help_text="Start time of the first segment in track"),
                ),
                (
                    "start_latitude",
                    models.FloatField(
                        editable=False, help_text="Latitude of the first recorded point from the *.gpx file"
                    ),
                ),
                (
                    "start_longitude",
                    models.FloatField(
                        editable=False, help_text="Longitude of the first recorded point from the *.gpx file"
                    ),
                ),
                ("start_temperature", models.FloatField(blank=True, help_text="Temperature at start.", null=True)),
                (
                    "start_weather_state",
                    models.CharField(blank=True, help_text="Weather state at start.", max_length=127, null=True),
                ),
                (
                    "short_start_address",
                    models.CharField(
                        blank=True, help_text="The short address of the start point", max_length=255, null=True
                    ),
                ),
                (
                    "full_start_address",
                    models.CharField(
                        blank=True, help_text="The full address of the start point", max_length=255, null=True
                    ),
                ),
                (
                    "finish_time",
                    models.DateTimeField(editable=False, help_text="End time of the last segment in track"),
                ),
                ("finish_latitude", models.FloatField(editable=False, help_text="Latitude of the finish point")),
                ("finish_longitude", models.FloatField(editable=False, help_text="Longitude of the finish point")),
                ("finish_temperature", models.FloatField(blank=True, help_text="Temperature at finish.", null=True)),
                (
                    "finish_weather_state",
                    models.CharField(blank=True, help_text="Weather state at finish.", max_length=127, null=True),
                ),
                (
                    "short_finish_address",
                    models.CharField(
                        blank=True, help_text="The short address of the finish point", max_length=255, null=True
                    ),
                ),
                (
                    "full_finish_address",
                    models.CharField(
                        blank=True, help_text="The full address of the finish point", max_length=255, null=True
                    ),
                ),
                ("points_no", models.PositiveIntegerField(blank=True, help_text="Number of points in GPX", null=True)),
                (
                    "length",
                    models.PositiveIntegerField(
                        blank=True,
                        help_text="Length in meters (calculated 3-dimensional used latitude, longitude, and elevation)",
                        null=True,
                    ),
                ),
                ("duration", models.PositiveIntegerField(blank=True, help_text="Duration in seconds", null=True)),
                (
                    "net_duration",
                    models.TimeField(
                        blank=True,
                        help_text="The officially measured time (e.g. from the official timekeeping of a running event.)",
                        null=True,
                        verbose_name="Net Duration",
                    ),
                ),
                (
                    "pace",
                    models.DecimalField(
                        blank=True,
                        decimal_places=2,
                        help_text="Min/km (number of minutes it takes to cover a kilometer)",
                        max_digits=4,
                        null=True,
                    ),
                ),
                ("uphill", models.IntegerField(blank=True, help_text="Uphill elevation climbs in meters", null=True)),
                (
                    "downhill",
                    models.IntegerField(blank=True, help_text="Downhill elevation descent in meters", null=True),
                ),
                ("min_elevation", models.IntegerField(blank=True, help_text="Minimum elevation in meters", null=True)),
                ("max_elevation", models.IntegerField(blank=True, help_text="Maximum elevation in meters", null=True)),
                (
                    "heart_rate_min",
                    models.PositiveIntegerField(
                        blank=True, editable=False, help_text="Minimum heart rate.", null=True
                    ),
                ),
                (
                    "heart_rate_avg",
                    models.PositiveIntegerField(
                        blank=True, editable=False, help_text="Average heart rate.", null=True
                    ),
                ),
                (
                    "heart_rate_max",
                    models.PositiveIntegerField(
                        blank=True, editable=False, help_text="Maximum heart rate.", null=True
                    ),
                ),
                (
                    "event",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="tracks",
                        to="for_runners.EventModel",
                    ),
                ),
                (
                    "ideal_distance",
                    models.ForeignKey(
                        blank=True,
                        help_text="Length in meters (calculated 3-dimensional used latitude, longitude, and elevation)",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="tracks",
                        to="for_runners.DistanceModel",
                    ),
                ),
                (
                    "lastupdateby",
                    models.ForeignKey(
                        blank=True,
                        editable=False,
                        help_text="User as last edit this entry",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="gpxmodel_lastupdateby",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "tracked_by",
                    models.ForeignKey(
                        blank=True,
                        editable=False,
                        help_text="The user that tracked this gpx entry",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="gpxmodel_createby",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "GPX Track",
                "verbose_name_plural": "GPX Tracks",
                "ordering": ("-start_time", "-pk"),
            },
        ),
        migrations.AddField(
            model_name="eventlinkmodel",
            name="event",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, related_name="links", to="for_runners.EventModel"
            ),
        ),
        migrations.AlterUniqueTogether(
            name="gpxmodel",
            unique_together={
                (
                    "start_time",
                    "start_latitude",
                    "start_longitude",
                    "finish_time",
                    "finish_latitude",
                    "finish_longitude",
                )
            },
        ),
    ]



"""
    created 17.11.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import csv
import logging


from django.core.files.base import ContentFile

# https://github.com/jedie/django-for-runners
from for_runners.svg import gpx2svg_string


log = logging.getLogger(__name__)


def generate_svg(gpx_track, force=False):
    """
    Generate SVG and save it to the given GPXModel.

    :param gpx_track: for_runners.models.gpx.GpxModel instance
    :param gpxpy_instance: gpxpy.parser.GPXParser instance
    :param force: don't check if svg track already exists
    :return: svg
    """
    log.debug("Create SVG from GPX...")

    if gpx_track.track_svg.name:
        # svg image already exists.
        if not force:
            log.info("Don't recreate existing images")
            return

    log.debug("Create SVG from GPX...")
    gpxpy_instance = gpx_track.get_gpxpy_instance()
    svg_string = gpx2svg_string(gpxpy_instance)
    content = ContentFile(svg_string)

    # https://docs.djangoproject.com/en/2.0/ref/models/fields/#django.db.models.fields.files.FieldFile.save
    gpx_track.track_svg.save(
        name="temp.svg", content=content, save=False  # real file path will be set in self.get_svg_upload_path()
    )
    log.debug(f"SVG created: {gpx_track.track_svg!r}")
    gpx_track.save()


class CsvGenerator:
    """
    TODO: Replace with for_runners.admin.gpx_import_export.GpxModelResource
    """

    HEADER_DATE = "date"
    HEADER_NAME = "name"
    HEADER_EVENT = "event"
    HEADER_USER = "user"
    HEADER_LENGTH = "length (km)"
    HEADER_DURATION = "duration"
    HEADER_PACE = "pace"
    HEADER_HEART_RATE = "heart rate"
    HEADER_TEMPERATURE = "temperature"
    HEADER_WEATHER = "weather"
    HEADER_CREATOR = "creator"

    def __init__(self, *, csv_file, add_username=True):
        """
        :param csv_file: file object, opened for writing
        :param add_username: add the username as csv column
        """
        self.csv_file = csv_file
        self.add_username = add_username

        fieldnames = [
            self.HEADER_DATE,
            self.HEADER_NAME,
            self.HEADER_EVENT,
            self.HEADER_LENGTH,
            self.HEADER_DURATION,
            self.HEADER_PACE,
            self.HEADER_HEART_RATE,
            self.HEADER_TEMPERATURE,
            self.HEADER_WEATHER,
            self.HEADER_CREATOR,
        ]
        if add_username:
            fieldnames.insert(3, self.HEADER_USER)

        self.csv_writer = csv.DictWriter(csv_file, fieldnames=fieldnames)
        self.csv_writer.writeheader()

    def add_gpx_track(self, track):
        if track.heart_rate_avg:
            heart_rate = "%i b/m" % track.heart_rate_avg
        else:
            heart_rate = ""

        if track.start_temperature:
            temperature = "%i°C" % round(track.start_temperature, 1)
            weather = track.start_weather_state
        else:
            temperature = ""
            weather = ""

        row = {
            self.HEADER_DATE: track.start_time.isoformat(),
            self.HEADER_NAME: track.short_name(start_time=False),
            self.HEADER_EVENT: "x" if track.participation else "",
            self.HEADER_LENGTH: round(track.length / 1000, 2),
            self.HEADER_DURATION: track.human_duration(),
            self.HEADER_PACE: track.human_pace(),
            self.HEADER_HEART_RATE: heart_rate,
            self.HEADER_TEMPERATURE: temperature,
            self.HEADER_WEATHER: weather,
            self.HEADER_CREATOR: track.creator,
        }
        if self.add_username:
            row[self.HEADER_USER] = track.tracked_by.username

        self.csv_writer.writerow(row)



# Generated by Django 2.0.7 on 2018-07-31 18:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("for_runners", "0004_auto_20180719_1739")]

    operations = [
        migrations.RemoveField(model_name="gpxmodel", name="net_duration"),
        migrations.AlterField(
            model_name="gpxmodel",
            name="duration",
            field=models.PositiveIntegerField(
                blank=True, help_text="Duration in seconds (from the GPX track)", null=True
            ),
        ),
        migrations.AlterField(
            model_name="gpxmodel",
            name="length",
            field=models.PositiveIntegerField(
                blank=True,
                help_text="Length in meters (calculated from GPX track 3-dimensional used latitude, longitude, and elevation)",
                null=True,
            ),
        ),
        migrations.RenameField(model_name="gpxmodel", old_name="duration", new_name="duration_s"),
        migrations.AlterField(
            model_name="gpxmodel",
            name="duration_s",
            field=models.PositiveIntegerField(
                blank=True, editable=False, help_text="Duration in seconds (from the GPX track)", null=True
            ),
        ),
    ]



# Generated by Django 2.1.8 on 2019-04-02 16:39

from django.db import migrations

# https://github.com/jedie/django-for-runners
from for_runners_project.cli import call_manage_command


def forwards(apps, schema_editor):
    from for_runners.management.commands import recreate_files

    call_manage_command(cmd_class=recreate_files)


class Migration(migrations.Migration):

    dependencies = [("for_runners", "0008_auto_20190402_1731")]

    operations = [migrations.RunPython(forwards, migrations.RunPython.noop)]



"""
    created 02.04.2019 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018-2019 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import logging
import statistics

# https://github.com/jedie/django-for-runners
from for_runners.geo import reverse_geo
from for_runners.gpx import get_extension_data, get_identifier
from for_runners.models import DistanceModel
from for_runners.weather import NoWeatherData, meta_weather_com


log = logging.getLogger(__name__)


def calculate_values(*, gpx_track):
    if not gpx_track.gpx:
        return

    gpxpy_instance = gpx_track.get_gpxpy_instance()
    gpx_track.points_no = gpxpy_instance.get_points_no()
    gpx_track.length = gpxpy_instance.length_3d()

    try:
        ideal_distances_qs = DistanceModel.objects.filter(
            min_distance_m__lte=gpx_track.length, max_distance_m__gte=gpx_track.length
        )
    except DistanceModel.DoesNotExist:
        pass
    else:
        ideal_distance_count = ideal_distances_qs.count()
        if ideal_distance_count > 1:
            log.error("Found more the one ideal distances for %i Meters", gpx_track.length)
            ideal_distances_qs = ideal_distances_qs.order_by("distance_km")
            ideal_distance = ideal_distances_qs[0]
        elif ideal_distance_count == 1:
            ideal_distance = ideal_distances_qs.get()
        else:
            ideal_distance = None

        if ideal_distance:
            gpx_track.ideal_distance = ideal_distance
            log.debug("Set ideal distance to %s", gpx_track.ideal_distance)

    # e.g: GPX without a track return 0
    duration = gpxpy_instance.get_duration()
    if duration:
        gpx_track.duration_s = duration
        gpx_track.calc_pace()

    uphill_downhill = gpxpy_instance.get_uphill_downhill()
    gpx_track.uphill = uphill_downhill.uphill
    gpx_track.downhill = uphill_downhill.downhill

    elevation_extremes = gpxpy_instance.get_elevation_extremes()
    gpx_track.min_elevation = elevation_extremes.minimum
    gpx_track.max_elevation = elevation_extremes.maximum

    identifier = get_identifier(gpxpy_instance)

    gpx_track.start_time = identifier.start_time
    gpx_track.finish_time = identifier.finish_time
    gpx_track.start_latitude = identifier.start_lat
    gpx_track.start_longitude = identifier.start_lon
    gpx_track.finish_latitude = identifier.finish_lat
    gpx_track.finish_longitude = identifier.finish_lon

    if not gpx_track.start_temperature:
        try:
            temperature, weather_state = meta_weather_com.coordinates2weather(
                gpx_track.start_latitude,
                gpx_track.start_longitude,
                date=gpx_track.start_time,
                max_seconds=gpx_track.duration_s,
            )
        except NoWeatherData:
            log.error("No weather data for start.")
        else:
            gpx_track.start_temperature = temperature
            gpx_track.start_weather_state = weather_state

    if not gpx_track.finish_temperature:
        try:
            temperature, weather_state = meta_weather_com.coordinates2weather(
                gpx_track.finish_latitude,
                gpx_track.finish_longitude,
                date=gpx_track.finish_time,
                max_seconds=gpx_track.duration_s,
            )
        except NoWeatherData:
            log.error("No weather data for finish.")
        else:
            gpx_track.finish_temperature = temperature
            gpx_track.finish_weather_state = weather_state

    if not gpx_track.full_start_address:
        try:
            start_address = reverse_geo(gpx_track.start_latitude, gpx_track.start_longitude)
        except Exception as err:
            # e.g.: geopy.exc.GeocoderTimedOut: Service timed out
            log.error(f"Can't reverse geo: {err}")
        else:
            gpx_track.short_start_address = start_address.short
            gpx_track.full_start_address = start_address.full

    if not gpx_track.full_finish_address:
        try:
            finish_address = reverse_geo(gpx_track.finish_latitude, gpx_track.finish_longitude)
        except Exception as err:
            # e.g.: geopy.exc.GeocoderTimedOut: Service timed out
            log.error(f"Can't reverse geo: {err}")
        else:
            gpx_track.short_finish_address = finish_address.short
            gpx_track.full_finish_address = finish_address.full

    # TODO: Handle other extensions, too.
    # Garmin containes also 'cad'
    extension_data = get_extension_data(gpxpy_instance)
    if extension_data is not None and "hr" in extension_data:
        heart_rates = extension_data["hr"]
        gpx_track.heart_rate_min = min(heart_rates)
        gpx_track.heart_rate_avg = statistics.median(heart_rates)
        gpx_track.heart_rate_max = max(heart_rates)

    if not gpx_track.creator:
        gpx_track.creator = gpxpy_instance.creator



# Generated by Django 2.1.8 on 2019-04-02 16:28

from django.db import migrations, models
import django_tools.file_storage.file_system_storage
import for_runners.models.gpx


class Migration(migrations.Migration):

    dependencies = [
        ('for_runners', '0006_auto_20180902_1958'),
    ]

    operations = [
        migrations.AddField(
            model_name='gpxmodel',
            name='gpx_file',
            field=models.FileField(blank=True, max_length=511, null=True, storage=django_tools.file_storage.file_system_storage.OverwriteFileSystemStorage(create_backups=True), upload_to=for_runners.models.gpx.gpx_upload_path, verbose_name='GPX Track'),
        ),
        migrations.AlterField(
            model_name='gpxmodel',
            name='track_svg',
            field=models.FileField(blank=True, null=True, storage=django_tools.file_storage.file_system_storage.OverwriteFileSystemStorage(create_backups=False), upload_to=for_runners.models.gpx.svg_upload_path, verbose_name='Track SVG'),
        ),
    ]



import unittest

import pytest
# https://github.com/jedie/django-tools
from django_tools.unittest_utils.selenium_utils import (
    SeleniumChromiumTestCase,
    SeleniumFirefoxTestCase,
    chromium_available,
    firefox_available,
)
from django_tools.unittest_utils.unittest_base import BaseTestCase
from django_tools.unittest_utils.user import TestUserMixin

# https://github.com/jedie/django-for-runners
from for_runners import __version__


@pytest.mark.django_db
class AdminAnonymousTests(BaseTestCase):
    """
    Anonymous will be redirected to the login page.
    """

    def test_login_en(self):
        response = self.client.get("/en/admin/", HTTP_ACCEPT_LANGUAGE="en")
        self.assertRedirects(response, expected_url="/en/admin/login/?next=/en/admin/")

    def test_login_de(self):
        response = self.client.get("/de/admin/", HTTP_ACCEPT_LANGUAGE="de")
        self.assertRedirects(response, expected_url="/de/admin/login/?next=/de/admin/")


@pytest.mark.django_db
class AdminLoggedinTests(TestUserMixin, AdminAnonymousTests):
    """
    Some basics test with the django admin
    """

    def test_staff_admin_index(self):
        self.login(usertype="staff")
        response = self.client.get("/en/admin/", HTTP_ACCEPT_LANGUAGE="en")
        self.assertResponse(
            response,
            must_contain=(
                f"<title>Site administration | Django-ForRunners v{__version__}</title>",
                "<h1>Site administration</h1>",
                "<strong>staff_test_user</strong>",
                "<p>You don't have permission to view or edit anything.</p>",
            ),
            must_not_contain=("error", "traceback", "/add/", "/change/"),
            template_name="admin/index.html",
            messages=[],
            html=True,
        )

    def test_superuser_admin_index(self):
        self.login(usertype="superuser")
        response = self.client.get("/en/admin/", HTTP_ACCEPT_LANGUAGE="en")
        self.assertResponse(
            response,
            must_contain=(
                "Django-ForRunners",
                "superuser",
                "Site administration",
                "/admin/auth/group/add/",
                "/admin/auth/user/add/",
            ),
            must_not_contain=("error", "traceback"),
            template_name="admin/index.html",
        )


@unittest.skipUnless(chromium_available(), "Skip because Chromium is not available!")
class AdminChromiumTests(SeleniumChromiumTestCase):
    def test_admin_login_page(self):
        self.driver.get(self.live_server_url + "/admin/login/")
        self.assert_equal_page_title(f"Log in | Django-ForRunners v{__version__}")
        self.assert_in_page_source('<form action="/en/admin/login/" method="post" id="login-form">')
        self.assert_no_javascript_alert()


@unittest.skipUnless(firefox_available(), "Skip because Firefox is not available!")
class AdminFirefoxTests(SeleniumFirefoxTestCase):
    def test_admin_login_page(self):
        self.driver.get(self.live_server_url + "/admin/login/")
        self.assert_equal_page_title(f"Log in | Django-ForRunners v{__version__}")
        self.assert_in_page_source('<form action="/en/admin/login/" method="post" id="login-form">')
        self.assert_no_javascript_alert()



from pathlib import Path
from unittest.mock import patch

from django.core.management import call_command
from django.test import TestCase
# https://github.com/jedie/django-tools
from django_tools.unittest_utils.assertments import assert_is_dir, assert_pformat_equal
from django_tools.unittest_utils.stdout_redirect import StdoutStderrBuffer
from django_tools.unittest_utils.user import TestUserMixin

# https://github.com/jedie/django-for-runners
import for_runners
from for_runners.models import GpxModel


# TODO:
# class ImportCommandTestCase(ForRunnersCommandTestCase):
#     def test_in_help(self):
#         output = self._call_manage(["--help"])
#         print(output)
#         self.assertIn("Available subcommands:", output)
#         self.assertIn("[for_runners]", output)
#         self.assertIn("import_gpx", output)
#
#     def test_import_help(self):
#         output = self._call_manage(["import_gpx", "--help"])
#         print(output)
#         self.assertIn("usage: manage import_gpx", output)
#         self.assertIn("Path to *.gpx files", output)
#
#     def test_import_no_username_given(self):
#         output = self._call_manage(["import_gpx", "/foo/bar"], excepted_exit_code=2)
#         print(output)
#         self.assertIn("following arguments are required: --username", output)
#
#     def test_import_wrong_username(self):
#         output = self._call_manage(
#             ["import_gpx", "--username", "NotExistingUser", "/foo/bar"], excepted_exit_code=3
#         )
#         print(output)
#         self.assertIn("ERROR getting user 'NotExistingUser'", output)
#         self.assertIn("Existing usernames are", output)
#         self.assertIn("test", output)
#
#     def test_import_wrong_path(self):
#         output = self._call_manage(
#             ["import_gpx", "--username", "test", "/foo/bar"], excepted_exit_code=4
#         )
#         print(output)
#         self.assertIn("ERROR: Given path '/foo/bar' is not a existing directory!", output)


class ImportTestCase(TestUserMixin, TestCase):
    def test_import(self):
        test_username = "normal_test_user"

        assert GpxModel.objects.filter(tracked_by__username=test_username).count() == 0

        base_path = Path(for_runners.__file__).parent
        fixture_files_path = Path(base_path, "tests/fixture_files")
        assert_is_dir(fixture_files_path)

        with StdoutStderrBuffer() as buff:
            with patch(
                "for_runners.services.gpx_calculate_values.meta_weather_com.coordinates2weather",
                return_value=(27.94, "Light Cloud"),
            ):
                call_command("import_gpx", "--username", test_username, str(fixture_files_path))

        output = buff.get_output()
        print(output)

        self.assertIn("Add new gpx tracks for user: normal_test_user", output)
        self.assertIn("1 - Add new track: 2018-02-21 Moers", output)
        self.assertIn("2 - Add new track: 2011-01-13 Berlin Tiergarten", output)
        self.assertIn("Added 2 new gpx tracks.", output)
        self.assertIn("User normal_test_user has now 2 tracks.", output)

        qs = GpxModel.objects.filter(tracked_by__username=test_username)

        existing_tracks = [str(track) for track in qs]
        assert_pformat_equal(
            existing_tracks, ["2018-02-21 Moers", "2011-01-13 Berlin Tiergarten"]
        )

        assert qs.count() == 2



from for_runners.gpx import cutted_hash


def test_cutted_hash():
    assert cutted_hash("foobar", length=20) == "BJICMHV5DI4Q73JL6MTP"
    assert cutted_hash("foobar", length=10) == "BJICMHV5DI"
    assert cutted_hash("foobar", length=4) == "BJIC"



import os
import subprocess
import unittest
from pathlib import Path

from for_runners import __version__
from for_runners_project import cli


def subprocess_output(*args, **kwargs):
    kwargs["stderr"] = subprocess.STDOUT
    print("_" * 100)
    print("Call:", args, kwargs)
    try:
        output = subprocess.check_output(*args, **kwargs)
    except subprocess.CalledProcessError as err:
        print(err.output.decode("UTF-8"))
        raise

    output = output.decode("UTF-8")
    print(output)
    return output


def assert_is_file(path):
    assert path.is_file(), f"File '{path}' not exist!"


def assert_is_dir(path):
    assert path.is_dir(), f"Direcotry '{path}' not exist!"


class CLITest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cli_file_path = Path(cli.__file__)

    def test_cli_executable(self):
        assert_is_file(self.cli_file_path)
        self.assertTrue(os.access(str(self.cli_file_path), os.X_OK), f"File '{self.cli_file_path}' not executeable!")

    def test_version(self):
        output = subprocess_output([f"{self.cli_file_path}", "--version"])
        self.assertIn(f"version {__version__}", output)

    def test_help(self):
        output = subprocess_output([f"{self.cli_file_path}", "--help"])
        self.assertIn("Usage: cli.py [OPTIONS] COMMAND [ARGS]...", output)
        self.assertIn("create-starter", output)
        self.assertIn("run-dev-server", output)
        self.assertIn("run-server", output)



"""
    Django settings for test_project project.

    Good startpoint fot Django-CMS stuff is:

    https://github.com/nephila/djangocms-installer/blob/develop/djangocms_installer/config/settings.py


    For more information on this file, see
    https://docs.djangoproject.com/en/1.8/topics/settings/

    For the full list of settings and their values, see
    https://docs.djangoproject.com/en/1.8/ref/settings/
"""

import logging
import sys as __sys
from pathlib import Path as __Path

from debug_toolbar.settings import CONFIG_DEFAULTS as DEBUG_TOOLBAR_CONFIG
from django.utils.translation import ugettext_lazy as _
from django_tools.settings_utils import FnMatchIps

from for_runners.app_settings import *  # noqa
from for_runners_project.utils.venv import get_venv_path as __get_venv_path


print("Use settings:", __file__)


# Build paths inside the project:
BASE_PATH = __Path(__file__).resolve().parent

# VirtualEnv root dir, e.g.: /home/<username>/DjangoForRunnersEnv
VENV_PATH = __get_venv_path()

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.8/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "Only for the tests ;)"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

SITE_ID = 1

# Required for the debug toolbar to be displayed:
INTERNAL_IPS = FnMatchIps(["localhost", "127.0.0.1", "::1", "172.*.*.*", "192.168.*.*", "10.0.*.*"])

ALLOWED_HOSTS = INTERNAL_IPS

DATA_UPLOAD_MAX_MEMORY_SIZE = 5000000

# Application definition

INSTALLED_APPS = (
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.sites",
    "debug_toolbar",  # https://github.com/jazzband/django-debug-toolbar/
    "autotask",  # https://bitbucket.org/kbr/autotask
    # 'easy_thumbnails', # https://github.com/SmileyChris/easy-thumbnails
    # 'treebeard', # https://github.com/django-treebeard/django-treebeard
    # 'sekizai', # https://github.com/ojii/django-sekizai
    # 'djangocms_text_ckeditor', # https://github.com/divio/djangocms-text-ckeditor
    # 'filer', # https://github.com/divio/django-filer
    "dynamic_fixtures",  # https://github.com/Peter-Slump/django-dynamic-fixtures
    "import_export",  # https://github.com/django-import-export/django-import-export
    # Django-ForRunners
    "for_runners",
    "for_runners_project.for_runners_helper_app",
)

# work-a-round for https://bitbucket.org/kbr/autotask/pull-requests/3/
if "run_test_project_dev_server" in __sys.argv:
    # Activate autotask only if our server command runs
    # for_runners_project.for_runners_helper_app.management.commands.run_server.Command
    print("Set: 'AUTOTASK_IS_ACTIVE = True'")
    AUTOTASK_IS_ACTIVE = True
else:
    AUTOTASK_IS_ACTIVE = False


# Download map via geotiler in for_runners.gpx_tools.gpxpy2map.generate_map
MAP_DOWNLOAD = True
# MAP_DOWNLOAD = False

ROOT_URLCONF = "for_runners_project.urls"
WSGI_APPLICATION = "for_runners_project.wsgi.application"

MIDDLEWARE = (
    # https://github.com/jazzband/django-debug-toolbar/
    "debug_toolbar.middleware.DebugToolbarMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.locale.LocaleMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "django.middleware.security.SecurityMiddleware",
    "django_tools.middlewares.ThreadLocal.ThreadLocalMiddleware",
)

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [str(__Path(BASE_PATH, "templates/"))],
        "OPTIONS": {
            "loaders": [
                (
                    "django.template.loaders.cached.Loader",
                    ("django.template.loaders.filesystem.Loader", "django.template.loaders.app_directories.Loader"),
                )
            ],
            "context_processors": [
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "django.template.context_processors.i18n",
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.template.context_processors.media",
                "django.template.context_processors.csrf",
                "django.template.context_processors.tz",
                "django.template.context_processors.static",
                "for_runners.context_processors.for_runners_version_string",
            ],
        },
    }
]

if DEBUG:
    # Disable caches:
    CACHES = {"default": {"BACKEND": "django.core.cache.backends.dummy.DummyCache"}}
    # Disable CacheLoader:
    TEMPLATES[0]["OPTIONS"]["loaders"] = (
        "django.template.loaders.filesystem.Loader",
        "django.template.loaders.app_directories.Loader",
    )

# Database
# https://docs.djangoproject.com/en/1.8/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": str(__Path(VENV_PATH, "Django-ForRunners-database.sqlite3")),
        # 'NAME': ":memory:"
        # https://docs.djangoproject.com/en/dev/ref/databases/#database-is-locked-errors
        "timeout": 30,
    }
}
print(f"Use Database: {DATABASES['default']['NAME']!r}")

# Internationalization
# https://docs.djangoproject.com/en/1.11/topics/i18n/

# Default and fallback language:
# https://docs.djangoproject.com/en/1.11/ref/settings/#language-code
LANGUAGE_CODE = "en"

# http://django-parler.readthedocs.org/en/latest/quickstart.html#configuration
PARLER_LANGUAGES = {
    1: [
        {"name": _("German"), "code": "de", "fallbacks": [LANGUAGE_CODE], "hide_untranslated": False},
        {"name": _("English"), "code": "en", "fallbacks": ["de"], "hide_untranslated": False},
    ],
    "default": {"fallbacks": [LANGUAGE_CODE], "redirect_on_fallback": False},  # all SITE_ID"s
}

# https://docs.djangoproject.com/en/1.8/ref/settings/#languages
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGES = tuple([(d["code"], d["name"]) for d in PARLER_LANGUAGES[1]])

LANGUAGE_DICT = dict(LANGUAGES)  # useful to get translated name by language code

# http://django-parler.readthedocs.org/en/latest/quickstart.html#configuration
PARLER_DEFAULT_LANGUAGE_CODE = LANGUAGE_CODE

USE_I18N = True

USE_L10N = True

# https://docs.djangoproject.com/en/1.11/topics/i18n/timezones/
TIME_ZONE = "Europe/Paris"
USE_TZ = True

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.8/howto/static-files/

STATIC_URL = "/static/"
STATIC_ROOT = str(__Path(BASE_PATH, "static"))

MEDIA_URL = "/media/"
MEDIA_ROOT = str(__Path(VENV_PATH, "media"))  # e.g.: /home/<username>/DjangoForRunnersEnv/media/
print(f"Tracks stored in {MEDIA_ROOT!r}")


# don't load jquery from ajax.googleapis.com, just use django's version:
DEBUG_TOOLBAR_CONFIG["JQUERY_URL"] = STATIC_URL + "admin/js/vendor/jquery/jquery.min.js"

# Basic Django CMS settings

CMS_TEMPLATES = (
    ("base.html", "Basic Page"),
    # ('tests/plugin_test.html', "Add CMS-Plugin test template")
)
CMS_PERMISSION = True

# Basic Placeholder config

# from djangocms_text_ckeditor.cms_plugins import TextPlugin
CKEDITOR = "TextPlugin"

CMS_PLACEHOLDER_CONF = {
    None: {
        "name": _("Content"),
        # Don't define 'plugin' to enable all existing CMS plugins
        # 'plugins': [
        #     CKEDITOR,
        #     plugin_anchor_menu_constants.ANCHOR_PLUGIN_NAME,
        #     plugin_anchor_menu_constants.DROP_DOWN_ANCHOR_MENU_PLUGIN_NAME,
        # ],
        "default_plugins": [{"plugin_type": CKEDITOR, "values": {"body": "Lorem ipsum dolor sit amet"}}],
    }
}

# Disable some more panels that will slow down the page:
DEBUG_TOOLBAR_CONFIG["DISABLE_PANELS"].add("debug_toolbar.panels.sql.SQLPanel")
DEBUG_TOOLBAR_CONFIG["DISABLE_PANELS"].add("debug_toolbar.panels.cache.CachePanel")

# don't load jquery from ajax.googleapis.com, just use django's version:
DEBUG_TOOLBAR_CONFIG["JQUERY_URL"] = "/static/admin/js/vendor/jquery/jquery.min.js"

DEBUG_TOOLBAR_CONFIG["SHOW_COLLAPSED"] = True  # Show toolbar collapsed by default.

# _____________________________________________________________________________
# cut 'pathname' in log output

old_factory = logging.getLogRecordFactory()


def cut_path(pathname, max_length):
    if len(pathname) <= max_length:
        return pathname
    return f"...{pathname[-(max_length - 3):]}"


def record_factory(*args, **kwargs):
    record = old_factory(*args, **kwargs)
    record.cut_path = cut_path(record.pathname, 30)
    return record


logging.setLogRecordFactory(record_factory)

# -----------------------------------------------------------------------------

# https://docs.python.org/3/library/logging.html#logging-levels
LOGGING = {
    "version": 1,
    "disable_existing_loggers": True,
    "formatters": {
        "colored": {  # https://github.com/borntyping/python-colorlog
            "()": "colorlog.ColoredFormatter",
            "format": "%(log_color)s%(asctime)s %(levelname)8s %(cut_path)s:%(lineno)-3s %(message)s",
        }
    },
    "handlers": {"console": {"class": "colorlog.StreamHandler", "formatter": "colored"}},
    "loggers": {
        "": {"handlers": ["console"], "level": "DEBUG", "propagate": False},
        "matplotlib": {"handlers": ["console"], "level": "INFO", "propagate": False},
        "django": {"handlers": ["console"], "level": "INFO", "propagate": False},
        "django_tools": {"handlers": ["console"], "level": "DEBUG", "propagate": False},
        "for_runners": {"handlers": ["console"], "level": "DEBUG", "propagate": False},
    },
}



from unittest import TestCase

# https://github.com/jedie/django-for-runners
from for_runners.geo import reverse_geo


class GeoTests(TestCase):
    def test_reverse_geo(self):

        # first point from: for_runners/tests/fixture_files/garmin_connect_1.gpx
        lat = "51.43788929097354412078857421875"
        lon = "6.617012657225131988525390625"

        address = reverse_geo(lat, lon)
        assert address.short == 'Moers'
        assert address.full.endswith('Moers, Kreis Wesel, Nordrhein-Westfalen, 47447, Deutschland')
